#!/bin/sh

luks_format() {
	# from mkswu scripts/common.sh
	# modifies dev with new target
	local target="$1"
	[ -n "$dev" ] || error "\$dev must be set"
	command -v cryptsetup > /dev/null \
		|| error "cryptsetup must be installed"
	command -v caam-encrypt > /dev/null \
		|| error "caam-encrypt must be installed"

	local index offset
	case "$dev" in
	*p[0-9])
		index=${dev##*p}
		index=$((index-1))
		offset="$(((9*1024 + index*4)*1024))"
		;;
	*) error "LUKS only supported on *p[0-9] partitions (got $dev)" ;;
        esac

	mkdir -p /run/caam
	local KEYFILE=/run/caam/lukskey
	# lower iter-time to speed PBKDF phase up,
	# since our key is random PBKDF does not help
	# also, we don't need a 16MB header so make it as small as possible (1MB)
	# by limiting the maximum number of luks keys (3 here, same size with less)
	# key size is 112
	unshare -m sh -c "mount -t tmpfs tmpfs /run/caam \
		&& caam-keygen create ${KEYFILE##*/} ccm -s 32 \
		&& dd if=/dev/random of=$KEYFILE.luks bs=$((4096-112-16)) count=1 status=none \
		&& dd if=/dev/random of=$KEYFILE.iv bs=16 count=1 status=none \
		&& cat $KEYFILE.iv $KEYFILE.luks > $KEYFILE.toenc \
		&& caam-encrypt $KEYFILE.bb AES-256-CBC $KEYFILE.toenc $KEYFILE.enc \
		&& cat $KEYFILE.bb $KEYFILE.iv $KEYFILE.enc > $KEYFILE.mmc \
		&& { if ! [ \$(stat -c %s $KEYFILE.mmc) = 4096 ]; then \
			echo \"Bad key size \$(stat -c %s $KEYFILE.mmc)\"; false; \
		fi; } \
		&& cryptsetup luksFormat -q --key-file $KEYFILE.luks \
			--pbkdf pbkdf2 --iter-time 1 \
			--luks2-keyslots-size=768k \
			$dev \
		&& cryptsetup luksOpen --key-file $KEYFILE.luks \
			--allow-discards $dev $target \
		&& dd if=$KEYFILE.mmc of=$DISK bs=4k count=1 status=none \
			oflag=seek_bytes seek=$offset" \
		|| error "Could not create luks partition on $dev"

	dev="/dev/mapper/$target"
}

luks_unlock() {
        # modifies dev if unlocked
        local target="$1"
        [ -n "$dev" ] || error "\$dev must be set"

        if [ -e "/dev/mapper/$target" ]; then
                # already unlocked, use it
                dev="/dev/mapper/$target"
                return
        fi

        command -v cryptsetup > /dev/null \
                || return 0

        # not luks? nothing to do!
        cryptsetup isLuks "$dev" \
                || return 0

        command -v caam-decrypt > /dev/null \
                || return 0

        local index offset
        case "$dev" in
        *p*)
                # keys are stored in $rootdev as follow
                # 0MB        <GPT header and partition table>
                # 9MB        key for part 1
                # 9MB+4k     key for part 2
                # 9MB+(n*4k) key for part n+1
                # 10MB       first partition
                index=${dev##*p}
                index=$((index-1))
                offset="$(((9*1024 + index*4)*1024))"
                ;;
        *) error "LUKS only supported on *p* partitions" ;;
        esac

        mkdir -p /run/caam
        local KEYFILE=/run/caam/lukskey
        # use unshared tmpfs to not leak key too much
        # key is:
        # - 112 bytes of caam black key
        # - 16 bytes of iv followed by rest of key
        unshare -m sh -c "mount -t tmpfs tmpfs /run/caam \
                && dd if=$DISK of=$KEYFILE.mmc bs=4k count=1 status=none \
                        iflag=skip_bytes skip=$offset \
                && dd if=$KEYFILE.mmc of=$KEYFILE.bb bs=112 count=1 status=none \
                && dd if=$KEYFILE.mmc of=$KEYFILE.enc bs=4k status=none \
                        iflag=skip_bytes skip=112 \
                && caam-decrypt $KEYFILE.bb AES-256-CBC $KEYFILE.enc \
                        $KEYFILE.luks >/dev/null 2>&1 \
                && cryptsetup luksOpen --key-file $KEYFILE.luks \
                        --allow-discards $dev $target >/dev/null 2>&1" \
                || return

        dev="/dev/mapper/$target"
}

format_ext() {
	local dev="$1"
	local label="$2"

	[ -e "$dev" ] || error "No partition $dev!"
	[ -n "$ENCRYPT_USERFS" ] && luks_format "${dev##*/}"
	dd if=/dev/zero of="$dev" bs=1M count=1
	mkfs.ext4 -L "$label" "$dev" \
		|| error "mkfs ext4 failed"
}

copy_firm() {
	local dev="$1" sz xxh
	local firm=/mnt/firm.squashfs

	[ -e "/firm.squashfs" ] && firm="/firm.squashfs"

	if [ -e "$firm" ]; then
		[ -e "$dev" ] || error "No partition $dev!"
		[ -n "$ENCRYPT_USERFS" ] && luks_format "${dev##*/}"
		dd if="$firm" of="$dev" bs=1M \
			|| error "Copy of firmware failed"
	fi
	if [ -e "$firm.xxh" ]; then
		sz=$(stat -c "%s" "$firm")
		xxh=$(dd if="$dev" bs=1M count="$sz" iflag=count_bytes,direct status=none \
			| xxhsum)
		xxh=${xxh%% *}
		[ "$(cat "$firm.xxh")" = "$xxh" ] \
			|| error "firmware checksum does not match, expected $(cat "$firm.xxh"), got $xxh"
	fi
	luks_unlock "${dev##*/}"
	if unshare -m mount "$dev" /mnt >/dev/null 2>&1; then
		case "$dev" in
		/dev/mapper/*) firmware_dev="/dev/mapper/${DISKPART#/dev/}4";;
		*) firmware_dev="${DISKPART}4";;
		esac
	fi
}

format_app() {
	local dev="$1"
	local label="app"

	[ -e "$dev" ] || error "No partition $dev!"
	[ -n "$ENCRYPT_USERFS" ] && luks_format "${dev##*/}"
	dd if=/dev/zero of="$dev" bs=1M count=1 status=none
	mkfs.btrfs -L "$label" -m DUP -R free-space-tree "$dev" \
		|| error "mkfs btrfs failed"
	mount "$dev" /mnt
	btrfs subvolume create /mnt/tmp \
		|| error "subvolume creation failed"
	mkdir /mnt/boot_0 /mnt/boot_1
	btrfs subvolume create /mnt/volumes \
		|| error "subvolume creation failed"
	btrfs subvolume create /mnt/boot_0/containers_storage \
		|| error "subvolume creation failed"
	btrfs subvolume create /mnt/boot_0/volumes \
		|| error "subvolume creation failed"

	# work around podman being silly and throwing errors if the
	# store does not look right
	podman run --storage-opt additionalimagestore="" \
		--root /mnt/boot_0/containers_storage image list >/dev/null 2>&1
	rm -f /mnt/boot_0/containers_storage/libpod/bolt_state.db
	# yes it really needs that too...
	mkdir -p /mnt/boot_0/containers_storage/overlay-layers /mnt/boot_0/containers_storage/overlay-images
	touch /mnt/boot_0/containers_storage/overlay-images/images.lock
	touch /mnt/boot_0/containers_storage/overlay-layers/layers.lock
	btrfs property set /mnt/boot_0/containers_storage ro true

	# and prepare boot_1 as well just in case
	btrfs subvolume snapshot -r /mnt/boot_0/containers_storage /mnt/boot_1/containers_storage \
		|| error "subvolume creation failed"
	btrfs subvolume create /mnt/boot_1/volumes \
		|| error "subvolume creation failed"

	mountpoint -q /mnt/boot_0/containers_storage/overlay \
		&& umount /mnt/boot_0/containers_storage/overlay
	umount /mnt
}

create_partitions() {
	local part start size

	# We need a first non-trivial sgdisk call to "resize" the gpt,
	# otherwise partition creations fail saying there is not
	# enough space... Rerandomize disk/partition GUID.
	sgdisk -G "$DISK"

	size="$(lsblk -o SIZE -rn "$DISKPART"1)"
	sgdisk --new 2:0:+"$size" -c 2:rootfs_1 \
		--new 3:0:+50M -c 3:logs \
		--new 4:0:+200M -c 4:firm \
		--new 5:0:0 -c 5:app \
		"$DISK" \
		|| error "partitionning disk failed"
}

initialize_partitions() {
	# first partition is in use in firstboot case, use losetup -P to
	# recreate a new device with its own partition table.
	loop=$(losetup --show -f -P "$DISK")
	dd if=/dev/zero of="${loop}p2" bs=1M count=1 \
		|| error "Could not wipe old 2nd rootfs"
	format_ext "${loop}p3" logs
	copy_firm "${loop}p4"
	format_app "${loop}p5"
	losetup -d "$loop"
	loop=""

	# format gp partitions we use
	if [ -e "$DISK"gp0 ]; then
		mkfs.ext4 -F "$DISK"gp0
		# XXX copy licenses
	fi
	if [ -e "$DISK"gp1 ]; then
		mkfs.fat --mbr=n -n ATLOG "$DISK"gp1
	fi
}

ext_board_eeprom_read() {
	i2cget -y 3 0x50 "$1" b 2>/dev/null
}

ext_board_is_atmark_techno() {
	# check vendor id is 0x0001 on big endian
	[ "$(ext_board_eeprom_read 0x00)" = 0x00 ] \
	&& [ "$(ext_board_eeprom_read 0x01)" = 0x01 ]
}

ext_board_is_lte() {
	# vendor id must be atmark techno and
	# product id is 0x0001 on big endian
	ext_board_is_atmark_techno \
	&& [ "$(ext_board_eeprom_read 0x02)" = 0x00 ] \
	&& [ "$(ext_board_eeprom_read 0x03)" = 0x01 ]
}

update_version() {
	local component=$1
	local version=$2
	local vers_file="/mnt/etc/sw-versions"

	awk -vcomp="$component" -vvers="$version" \
		'$1 == comp { print comp, vers; found=1 }
		 $1 != comp { print }
		 END { if (!found) print comp, vers }' \
			 < "$vers_file" > "$vers_file.tmp" \
		&& mv "$vers_file.tmp" "$vers_file" \
		|| error "Could not set $component version in sw-versions"
}

# required environment variables
#  DISK = base block device e.g. /dev/mmcblk2 or /dev/vda
#  DISKPART = prefix for partitions e.g. /dev/mmcblk2p or /dev/vda
firstboot() {
	local boot boot_version firmware_dev=""
	[ "${DISK#/dev/mmcblk}" != "$DISK" ] && DISK="${DISK%p[0-9]}" || DISK="$DISKPART"

	create_partitions

	# self-destruct, and poweroff only if we really did something
	local rootdev="${DISKPART}1"
	[ -n "$ENCRYPT_ROOTFS" ] && rootdev="/dev/mapper/rootfs_0"
	mount "$rootdev" /mnt \
		|| error "Could not mount $rootdev"
	mount -o remount,rw /mnt \
		|| error "Could not make $rootdev rw"
	initialize_partitions
	case "$DISK" in
	/dev/mmcblk2)
		boot="cat /dev/mmcblk2boot0"
		if [ -e /dev/mmcblk2gp1 ] && ! grep -q /var/at-log /mnt/etc/fstab; then
			cat >> /mnt/etc/fstab <<'EOF'
/dev/mmcblk2gp1	/var/at-log			vfat	defaults			0 0
EOF
		fi
		;;
	/dev/mmcblk1)
		boot="dd if=/dev/mmcblk1 bs=1k skip=32 count=4096"
		sed -i -e "s:/dev/mmcblk2p:$DISKPART:" /mnt/etc/fstab
		sed -i -e "s:/dev/mmcblk2boot0:$DISK:" /mnt/etc/fw_env.config
		;;
	*)
		# no env for scsi/sata disks
		sed -i -e 's:/dev/mmcblk2p3:LABEL=logs:' -e 's:/dev/mmcblk2p5:LABEL=app:' /mnt/etc/fstab
		> /mnt/etc/fw_env.config
		rm -f /mnt/etc/runlevels/default/reset_bootcount
		;;
	esac

	# update versions. later steps expect version file to exit for updating
	touch /mnt/etc/sw-versions
	[ -e /mnt/etc/atmark-release ] \
		&& update_version base_os "$(cat /mnt/etc/atmark-release)"
	if [ -n "$boot" ]; then
		boot_version=$(eval "$boot" | strings |
			grep -m1 -oE '20[0-9]{2}.[0-1][0-9]-[0-9a-zA-Z.-]*')
		[ -n "$boot_version" ] && update_version boot "$boot_version"
	fi
	if [ -n "$ENCRYPT_ROOTFS" ]; then
		# cannot easily guess version from compressed boot image, use 1
		update_version boot_linux 1
	else
		sed -i -e '/^boot_linux /d' /mnt/etc/sw-versions \
			|| error "Could not update sw-versions"
	fi

	if grep -q uninit /mnt/etc/hwrevision; then
		local board
		board=$(dd if=/sys/bus/nvmem/devices/imx-ocotp0/nvmem bs=4 skip=56 count=1 status=none | xxd -p | head -c 4)
		case "$board" in
		c600) echo 'iot-g4-eva at1' > /mnt/etc/hwrevision;;
		c700) echo 'iot-g4-es1 at1' > /mnt/etc/hwrevision;;
		c800) echo 'iot-g4-es2 at1' > /mnt/etc/hwrevision;;
		c900) echo 'AGX4500 at1' > /mnt/etc/hwrevision;;
		cb00) echo 'iot-g4-es3 at1' > /mnt/etc/hwrevision;;
		esac
	fi

	rm -f /mnt/firm.squashfs /mnt/firm.squashfs.xxh
	sed -i -e "/\/opt\/firmware/d" /mnt/etc/fstab \
		|| error "Could not update fstab"
	if [ -n "$firmware_dev" ]; then
		cat >> /mnt/etc/fstab <<EOF \
			|| error "Could not update fstab"
$firmware_dev	/opt/firmware			squashfs defaults			0 0
EOF
	fi
	if [ -n "$ENCRYPT_USERFS" ]; then
		sed -i -e 's@dev/\(mmcblk[12]p[35]\)@dev/mapper/\1@' /mnt/etc/fstab \
			|| error "Could not update fstab"
	else
		sed -i -e 's@dev/mapper/\(mmcblk[12]p[35]\)@dev/\1@' /mnt/etc/fstab \
			|| error "Could not update fstab"
	fi

	# if rootfs was copied from ab1 we also need to fix a couple more settings
	# this is harmless if useless, already done or for sd boot
	sed -i -e 's/subvol=boot_1/subvol=boot_0/' /mnt/etc/fstab \
		|| error "Could not update fstab"
	sed -i -e 's/mmcblk2boot1/mmcblk2boot0/' /mnt/etc/fw_env.config \
		|| error "Could not update fw_env.config"
	sed -i -e '/^other_boot\(_linux\)\? /d' /mnt/etc/sw-versions \
		|| error "Could not update sw-versions"

	if ext_board_is_lte; then
		einfo "LTE extension board exists"
		echo "fdt_overlays=armadillo_iotg_g4-lte-ext-board.dtbo" > /mnt/boot/overlays.txt
		ln -s /etc/init.d/connection-recover /mnt/etc/runlevels/default/
		ln -s /etc/init.d/modemmanager /mnt/etc/runlevels/boot/
	fi

	rm -f /mnt/etc/init.d/firstboot-atmark
	rm -f /mnt/etc/runlevels/sysinit/firstboot-atmark
	rm -f /mnt/lib/rc/sh/functions-atmark.sh
	date +%s.%N > /mnt/etc/.rootfs_update_timestamp
	umount /mnt
}

[ -e /installer.conf ] && . /installer.conf
